#pragma once

#include <string>
#include <functional>
#include <cerrno>

#if defined(__linux__) || defined(__APPLE__)
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include <netdb.h>
#elif  defined(WIN32)
#pragma comment(lib, "Ws2_32.lib")
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <winsock2.h>
#include <ws2ipdef.h>
#include <ws2tcpip.h>
#endif



#define FDR_UNUSED(expr){ (void)(expr); }
#define FDR_ON_ERROR std::function<void(int, std::string)> onError = [] (int errorCode, std::string errorMessage) {FDR_UNUSED(errorCode); FDR_UNUSED(errorMessage) }

#ifndef AS_DEFAULT_BUFFER_SIZE
#define AS_DEFAULT_BUFFER_SIZE 0x1000 // 4096 bytes
#endif

class BaseSocket
{
public:
    enum SocketType {
        TCP = SOCK_STREAM,
        UDP = SOCK_DGRAM
    };
    sockaddr_in address;

    void Close() {
#ifdef WIN32

#elif __linux__
        shutdown(this->sock, SHUT_RDWR);
        close(this->sock);
#endif
    }

    std::string RemoteAddress() const { return IPToString(this->address); };
    int RemotePort() const { return ntohs(this->address.sin_port); }
    int FileDescriptor() const { return this->sock; }
protected:
    int sock = 0;

    // Get std::string value of the IP from a `sockaddr_in` address struct
    static std::string IPToString(const sockaddr_in& addr)
    {
        char ip[INET_ADDRSTRLEN];

        inet_ntop(AF_INET, &(addr.sin_addr), ip, INET_ADDRSTRLEN);

        return std::string(ip);
    }

    BaseSocket(FDR_ON_ERROR, SocketType sockType = TCP, int socketID = -1)
    {
        if (socketID == -1) {
            this->sock = socket(AF_INET, sockType, 0);

            if (this->sock == -1)
            {
                onError(errno, "Socket creating error.");
            }
        } else {
            this->sock = socketID;
        }
    }

    virtual ~BaseSocket() {}
};