#pragma once

#include <Sockets/Commons.hpp>
#include <Sockets/Sockets.hpp>
#include <cstring>
#include <array>
#include <string>
#include <vector>
#ifdef __linux__
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>
#elif WIN32
#include <winsock2.h>

#endif


#include <Sockets/Exceptions.hpp>

namespace Socket
{
	typedef struct sockaddr_in sockaddr_in_t;
	typedef struct sockaddr sockaddr_t;
	struct IPAddress
	{
		std::array<u8, 4> octets{};
		u16 port{};
	public:

		static IPAddress Any(u16 portno) 		{ return IPAddress{INADDR_ANY, portno};}
		static IPAddress Loopback(u16 portno)  { return IPAddress{INADDR_LOOPBACK, portno};}
		static IPAddress Broadcast(u16 portno) { return IPAddress{INADDR_BROADCAST, portno};}

        // Tries to resolve an IP address from a URL
        static IPAddress Resolve(std::string const& uri, u16 port);

		IPAddress() {}
		IPAddress(const std::string& ipaddr, u16 port);
		IPAddress(u8 a, u8 b, u8 c, u8 d, u16 port);
		IPAddress(const sockaddr_in_t& addr_in);		
		explicit operator sockaddr_in_t() const;
		const u8& operator[](size_t octet) const;
		bool operator == (const IPAddress& other) const;
		bool operator != (const IPAddress& other) const;
		[[nodiscard]] std::string addr_string() const;
		[[nodiscard]] std::string port_string() const;
		[[nodiscard]] std::string to_string() const;
		operator std::string() const { return this->to_string();}
	private:
		IPAddress(uint32_t ipaddr, uint16_t portno);
	};


}
