#pragma once

#include <iostream>
#include <string>
#ifdef __linux__
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>
#elif WIN32

#endif

#include <memory>
#include <cstring>

#include <Sockets/Sockets.hpp>

#include <Sockets/IPAddress.hpp>

#ifndef INPORT_ANY
#define INPORT_ANY 0
#endif

namespace Socket {
    /// A general purpose TCP interface
    /// TCP (Transmission Control Protocol) provides reliable, ordered, and error-checked delivery of
    /// a stream of bytes between peers.
	class TcpSocket_Old {
	public:
		TcpSocket_Old();
		TcpSocket_Old(int family, int flags);
		TcpSocket_Old(int socket, addrinfo info, bool connected, bool bound);
		virtual ~TcpSocket_Old();

		TcpSocket_Old(const TcpSocket_Old &socket) = default;
		TcpSocket_Old &operator=(const TcpSocket_Old &socket) = delete;
		void Bind(uint16_t port);
		void BindAny();
		void BindAny(uint16_t&portno);
		void Connect(const IPAddress& ipaddr);
		void Connect(std::string address, uint16_t port);
		void Listen(int maxQueue);
		std::shared_ptr<TcpSocket_Old> Accept();
		void Send(const u8 *data, unsigned int length, int flags);
		// Receive data (blocking)
		// @return true if socket is still open, false otherwise
		bool Receive(u8* msg, int len, int flags);
		void Close();
	private:
		void setInfo(int port);
		void setInfo(std::string address, int port);
		void openSocket(addrinfo *info);
		addrinfo * mInfo;
		int mSock = -1;
		bool mSockCreated = false;
		bool mBound = false;
		bool mConnected = false;
		bool mClosed = false;
	};
}
